#!/usr/bin/perl

##################################################################
#IP-Generator v2 by bighawk                                      #
#                        oct 2000                                #
#                                                                #
#Try --help for possible options                                 #
#                                                                #
#questions or comments? mail bighawk@warfare.com                 #
#for all those annoying programs requesting an ip-list as input..#
#                                                                #
#Read the README file included in ipgenv2.tar.gz  package        #
#                                                                #
##################################################################

use Getopt::Long;

my ($ip_pattern, $rand, $endrand, $childs, $outfile, $res, $san, $nofork, $MAX_CHILDS);

#handling the options
GetOptions('help|h' => sub { &help }, 'resolve|r' => \$res, 'resolved-only|q' => sub { $res = 2 }, 'no-fork|n' => \$nofork,
           'outfile|o=s' => \$outfile,  'pattern|p=s' => \$ip_pattern, 'bypass-sanchk|b' => \$san, 'random|d=i' => \$rand,
           'end-rand|e=i' => \$endrand, 'childs|c=i' => \$MAX_CHILDS);

$MAX_CHILDS = 10 unless($MAX_CHILDS);                     # if MAX_CHILDS has not been specified, default of 10 is set
$outfile = "&STDOUT" unless($outfile);                    # if outfile has not been specified, STDOUT is set.            

die "Error: Couldn't open outfile.\n" unless (open(LOG, ">$outfile"));

if ($rand) {                                              # starting of --random block
   die "Error: the number of random ips to generate must be positive.\n" if ($rand < 0);
   die "Error: the endrand value must be between 0 and 255.\n" if($endrand < 0 || $endrand > 255);
   srand;                                                 # rand() seed initialisation
   for (1..($rand)) {
      $addr[0] = int(rand(254)+1);
      $addr[1] = int(rand(255));
      $addr[2] = int(rand(255));
      $addr[3] = int(rand(255));
      $addr[3] = $endrand if ($endrand);
      $b_ip = "$addr[0].$addr[1].$addr[2].$addr[3]";
      &printip;                                           # call to subroutine printip() wich will print the ips.
   }
   &waitchld;                                             # call to subroutine waitchld() wich will wait for childs to finish 
   close(LOG);
   print "\nDone\n" && exit(0);                           # end of --rand block and program
}

die "Error: not enough parameters. Try --help\n" unless ($ip_pattern);
die "Error: unsupported pattern. Try --help.\n" unless ($ip_pattern =~ /[0-9\-,*]+\.[0-9\-,*]+\.[0-9\-,*]+\.[0-9\-,*]+/);

($ip_part1, $ip_part2, $ip_part3, $ip_part4) = split(/\./, $ip_pattern);



@part1 = &checkpart("$ip_part1");    #
@part2 = &checkpart("$ip_part2");    # send each part of the ip-pattern to subroutine checkpart()
@part3 = &checkpart("$ip_part3");    # which returns the list of matched numbers.
@part4 = &checkpart("$ip_part4");    #

for ($a = 0; $a < @part1; $a++) {                                       #
   for ($b = 0; $b < @part2; $b++) {                                    #
      for ($c = 0; $c < @part3; $c++) {                                 # nested loop generating the ips
         for ($d = 0; $d < @part4; $d++) {                              # 
            $b_ip = "$part1[$a].$part2[$b].$part3[$c].$part4[$d]";
            @addr = ($part1[$a], $part2[$b], $part3[$c], $part4[$d]);
            &printip;                                                   #call to subroutine printip() wich will print the ips
         }
      }
   }
}

&waitchld;

close(LOG);
print "\nDone\n";

sub checkpart {
   my ($part) = @_;
   my (@returnlist);
   my (@tmp);
   if ($part =~ /^\d+$/) {                       # check if $part is a normal digit
      &sanchk($part);                            # first call to subroutine sanchk().
      $returnlist[0] = $part;                    
   } elsif  ($part =~ /\*/) {                    # check if $part is an asterisk
      for ($a = 0; $a <= 255; $a++) {            # generating the 256 possibilities
         $returnlist[$a] = $a;
      }
   } elsif ($part =~ /\d+,\d+/) {                # check if $part matches the seperator ","
      @tmp = split(/\,/, $part);                 # split $part and assign to a temporary array
      my ($num);     
      for ($cnt = 0; $cnt < @tmp; $cnt++) {      # looping trough @tmp to see if it contains ranges (-)
         if ($tmp[$cnt] =~ /(\d+)-(\d+)/) {      # if it does contain a range, generating all possibilities
            &sanchk($1);
            &sanchk($2);
            for ($a = $1; $a <= $2; $a++) {
               $returnlist[$num] = $a;           # and assigning them to @returnlist
               $num++;
            }
         } else {                                # if not, assign each number to @returnlist
            &sanchk($tmp[$cnt]);
            $returnlist[$num] = $tmp[$cnt];
            $num++;
         }
      }

   } elsif ($part =~ /^(\d+)-(\d+)$/) {          # check if $part contains only a range
      my $b = 0;
      &sanchk($1);
      &sanchk($2);
      for ($a = $1; $a <= $2; $a++) {            # looping trough the range and generating all possibilities 
         $returnlist[$b] = $a;                   # and assign them to @returnlist
         $b++;
      }
   } else {                                      # if none matched, "pattern" can't be handled
      die "Error: Can not handle $part.\n";
   }
   @returnlist;                                  # return @returnlist
}

sub sanchk {                                              # Subroutine sanchk() checks if ip_parts are valid ip_parts.
   return if ($san == 1);                                 # return if --bypass-sanchk is specified.
   my ($part) =  @_;
   if ($part < 0 || $part > 255) {
      die "Error: ip-parts must be between 0 and 255\n";
   }
}

sub resolve {                                             # Subroutine resolve() will resolve the ip
   my (@addrbytes) = @_;
   my $paddr = pack ("C4", @addrbytes);                   # pack ip parts to a 4 bytes integer.
   my @host = gethostbyaddr($paddr, 2);                   # and resolving the packed address
   return $host[0];                                       # and return resolved host (if resolved)
}

sub help {                                                # start of subroutine help wich will display built-in help
   print "\t-= IP-generator v2 by bighawk =-\n";
   print "\nSyntax: $0 <options>\n";
   print "Supported options are:\n";
   print " --h --help : To print this help.\n";
   print "One of the following 2 options must be given.\n";
   print " --p --pattern <pattern> : Specify the pattern to generate ips with. Cannot be used with --d.\n";
   print "                           IP-generator has a powerfull range specification syntax.\n";
   print "                           Supported are:\n";
   print "                             * :For all 256 possibilities.\n";
   print "                             , :To seperate single numbers.\n";
   print "                             - :To specify number-ranges.\n";
   print "                           Example: 112.*.1-10,15,20-40.0,255\n";
   print " --d --random <num> : Generates <num> random valid ips. Cannot be used with --p.\n";
   print " --e --end-rand <num> : Specifies where the random generated ip should end in. Can only be used with --d.\n";
   print " --o --outfile <file> : Writes output to <file>. If omitted output will be written to STDOUT.\n";
   print " --r --resolve : Try to resolve generated ips. Cannot be used with --q.\n";
   print " --q --resolved-only : Only print resolved ips to outfile. Cannot be used with --r.\n";
   print "Warning: both resolving options can slow the process down.\n";
   print " --c --childs <num> : Specifies number of simultaneous processes used. If omitted the default of 10 childs is used\n";
   print " --n --no-fork : Dont create multiple threads. \n";
   print " --b --bypass-sanchk : To bypass the built-in ip-part sanity check.\n";
   print "Tip: You can use it as a \"ip-resolve-script\".\n\n";
   exit;
}

sub printip {                                          # start of multithreaded subroutine printip.
   if (($res) && ($nofork == 0)) {                     # checking if ip needs to be resolved and if --no-fork is set
      my $pid;                     
      $pid = fork();                                   # creation of a new child
      if($pid > 0) {                                   # if parent process execute next lines
         $childs++;
         if($childs >= $MAX_CHILDS) {
            $wait_ret=wait();                          # wait till one child finishes if MAX_CHILDS has been reached
            if($wait_ret > 0) {
               $childs--;
            }
         }
         next;
      } elsif(undef $pid) {
         print "Error: fork error\n";                  # if fork() fails
         exit(0);
      } else {                                         # part to be executed by the childs
         my $host;
         local $SIG{'ALRM'} = sub { exit(0); };        # killing child on timeout
         $host = &resolve(@addr);                      # and resolve the ip 
         exit(0) if ($host eq "" && $res == 2);        # exit child if not resolved and --resolved-only is set
         $host ? print LOG "$host\n" : print LOG "$b_ip\n";
         exit(0);
      }
   } elsif(($res > 0) && ($nofork == 1)) {                 # if --resolve* is set and --no-fork is set
      $host = &resolve(@addr);
      next if ($host eq "" && $res == 2);
      $host ? print LOG "$host\n" : print LOG "$b_ip\n";
   } else {                                            # if ip doesnt need to be resolved
      print LOG "$b_ip\n";
   }
}

sub waitchld {                                         # starting of subroutine waitchld wich will wait till all childs
   for(1..$childs){                                    # are finished
      my $wt = wait();
      if($wt == -1){
         redo;
      }
   }
}
